#include "xsiGeoUtilities.h"

#include <vector>

const void		GetPointRefFrame(ToolContext& in_ctxt,X3DObject& in_obj, Vertex& in_vert, CVector3& outX, CVector3& outY, CVector3& outZ)
{
	
	bool boolvalid;

	outY=in_vert.GetNormal(boolvalid);
	//calc edge vector
	CEdgeRefArray eref = in_vert.GetNeighborEdges();
		
	Edge e(eref[0]);
	CVector3Array ep = e.GetPoints().GetPositionArray();
	CVector3 edgevector = ep[1];
	edgevector-=ep[0];
	edgevector.NormalizeInPlace();

	// outx = calc normalized projection of edgevector onto (outx) normal plane
	double d=edgevector.Dot(outY);
	CVector3 n=outY;
	n*=d;
	outX=edgevector;
	outX-=n;
	outX.NormalizeInPlace();
	outX.NegateInPlace();
	// z is cross of x,y
	outZ.Cross(outX,outY);	
	
}
const void		GetEdgeRefFrame(ToolContext& in_ctxt,X3DObject& in_obj, Edge& in_edge, CVector3& outX, CVector3& outY, CVector3& outZ)
{
	outY=getEdgeNormal(in_ctxt,in_edge,false);
	CVector3Array ep = in_edge.GetVertices().GetPositionArray();
	outX = ep[1];
	outX-= ep[0];
	outX.NormalizeInPlace();
	outY.NormalizeInPlace();
	outZ.Cross(outX,outY);	
	outZ.NormalizeInPlace();	
}
const void		GetPolyRefFrame(ToolContext& in_ctxt, X3DObject& in_obj, PolygonFace& in_poly, CVector3& outX, CVector3& outY, CVector3& outZ)
{	
	outY = getPolyNormal(in_ctxt,in_poly,false);
		
	CVector3Array ep = in_poly.GetVertices().GetPositionArray();
	CVector3 edgevector = ep[0];
	CVector3 edgesub=ep[1];
	edgesub-=edgevector;
	outX = edgesub;
	outX.NormalizeInPlace();	
	outZ.Cross(outX,outY);	
	outZ.NormalizeInPlace();
}
const CVector3  AlignPrimitiveToComponentRefFrame(ToolContext& in_ctxt, PolygonFace& in_poly,X3DObject& in_obj, CVector3& out_rotx,CVector3& out_roty,CVector3& out_rotz, bool switchaxes, bool objKineMult)
{
	out_rotx.PutNull();
	out_roty.PutNull();
	out_rotz.PutNull();
		
	GetPolyRefFrame(in_ctxt,in_obj, in_poly, out_rotx,out_roty,out_rotz);
		

	double rx,ry,rz;
	if( objKineMult )	{
		CMatrix3 m;
		m = in_obj.GetKinematics().GetGlobal().GetTransform().GetRotation().GetMatrix();
		out_rotx *= m;
		out_roty *= m;
		out_rotz *= m;
	}
	out_rotx.NormalizeInPlace();
	out_roty.NormalizeInPlace();
	out_rotz.NormalizeInPlace();
	//make sure vectors are orthogonal
	out_rotx.Cross( out_roty, out_rotz ); 
		
	out_rotx.NormalizeInPlace(); 
		 
	CRotation rot;	
	rot.SetFromXYZAxes(out_rotx,out_roty,out_rotz);
		
	rot.GetXYZAngles(rx,ry,rz);

	return CVector3(RadiansToDegrees(rx),RadiansToDegrees(ry),RadiansToDegrees(rz));
}

const CVector3 getEdgeNormal(ToolContext& in_ctxt,Edge &in_edge,bool in_mulmatrix)
{
		
	CPolygonFaceRefArray l_ep(in_edge.GetNeighborPolygons());
	int cnt = l_ep.GetCount();
	CVector3 vn;
	for (int i=0;i<cnt;i++)
	{
		PolygonFace face(l_ep[i]);
		CVector3 s,e,n;
		n=getPolyNormal(in_ctxt,face,false);
		n.NormalizeInPlace();
		vn+=n;
	}
	X3DObject in_obj=in_edge.GetParent();
	if(in_mulmatrix)
		vn*=in_obj.GetKinematics().GetGlobal().GetTransform().GetRotationMatrix3();
	vn.NormalizeInPlace();
	return vn;
}
const CVector3 getVertexNormal(ToolContext& in_ctxt,Vertex &in_v, bool in_mulmatrix)		//returns normal
{
	CPolygonFaceRefArray l_vp(in_v.GetNeighborPolygons());
	int cnt = l_vp.GetCount();
	CVector3 vn;

	for (int i=0;i<cnt;i++)
	{
		PolygonFace face(l_vp[i]);
		CVector3 s,e,n;
		n=getPolyNormal(in_ctxt, face,false);
		vn+=n;
	}
		
	X3DObject	in_obj=in_v.GetParent();
	if(in_mulmatrix)
		vn*=in_obj.GetKinematics().GetGlobal().GetTransform().GetRotationMatrix3();
	vn.NormalizeInPlace();
	return vn;
}
// calculate polygon normal (Newell's method) http://www.opengl.org/wiki/Calculating_a_Surface_Normal
const CVector3 getPolyNormal(ToolContext& in_ctxt, PolygonFace &in_poly, bool in_mulmatrix)
{
		X3DObject			in_obj = in_poly.GetParent();
		PolygonMesh			mesh( in_obj.GetActivePrimitive().GetGeometry() );
		CVertexRefArray		verts( in_poly.GetVertices());
		CLongArray			tri_ind( in_poly.GetTriangleSubIndexArray());
		CVector3Array		n,p;	
							
		UINT vertCnt = verts.GetCount();
		double x=0,y=0,z=0;
		for(UINT i=0;i<vertCnt;i++)
		{
			Vertex current = verts[i];			
			Vertex next = verts[ (i+1) % vertCnt];
			CVector3 c(current.GetPosition() );
			CVector3 n(next.GetPosition() );
			x= x+ ( (c[1]-n[1])* (c[2]+n[2]) );
			y= y+ ( (c[2]-n[2])* (c[0]+n[0]) );
			z= z+ ( (c[0]-n[0])* (c[1]+n[1]) );
		}

		CVector3 normal(x,y,z);		
		if( in_mulmatrix )
			normal *= in_obj.GetKinematics().GetGlobal().GetTransform().GetRotationMatrix3();
		normal.NormalizeInPlace();
		return normal;
}

const CStatus GetModelBoudingBox(X3DObject &in_obj, CVector3 &out_bboxmin, CVector3 &out_bboxmax, CVector3 &out_bboxhalfsize, bool useInputTransform, CTransformation in_ct )
{
	Model m( in_obj );

	if ( m.GetModelKind() == siModelKind_Reference )
		return CStatus::False;
	if ( m.GetModelKind() == siModelKind_Instance )
		m = m.GetInstanceMaster();
	CRefArray objs( m.GetChildren() );
	CVector3 bmin, bmax,hs;
	bool bboxset = false;

	for (int j=0;j <objs.GetCount(); j++)
	{
		if (objs[j].GetClassIDName() == L"X3DObject" || objs[j].GetClassIDName() == L"Model" )
		{
			X3DObject l_obj( objs[j] );

			if (l_obj.GetType() == L"polymsh")	{
				CTransformation obj_t (l_obj.GetKinematics().GetLocal().GetTransform() );
				obj_t.MulInPlace( in_ct );				
				GetBoudingBox(l_obj, bmin, bmax, hs, useInputTransform, obj_t );
			}
			else
				if (l_obj.GetType() == L"#model")	{
					CVector3 bmmax, bmmin;
					CTransformation obj_t (l_obj.GetKinematics().GetLocal().GetTransform() );
					obj_t.MulInPlace( in_ct );
					if ( GetModelBoudingBox(l_obj, bmmin, bmmax, hs, useInputTransform, obj_t ) == CStatus::OK )
					{
						bmin = bmmin;
						bmax = bmmax;
					}
					else {
						continue;
					}
				}
				else
					continue;

			if ( !bboxset ) //not set yet
			{
				out_bboxmin = bmin;
				out_bboxmax = bmax;
				bboxset = true;
			}
			else	{
				if ( bmin[0] < out_bboxmin[0] ) out_bboxmin.PutX( bmin[0] );
				if ( bmin[1] < out_bboxmin[1] ) out_bboxmin.PutY( bmin[1] );
				if ( bmin[2] < out_bboxmin[2] ) out_bboxmin.PutZ( bmin[2] );
				if ( bmax[0] > out_bboxmax[0] )	out_bboxmax.PutX( bmax[0] );
				if ( bmax[1] > out_bboxmax[1] )	out_bboxmax.PutY( bmax[1] );
				if ( bmax[2] > out_bboxmax[2] )	out_bboxmax.PutZ( bmax[2] );
			}
				
		}
	}
	return bboxset ? CStatus::OK : CStatus::False;
};

const void GetBoudingBox(X3DObject &in_obj, CVector3 &out_bboxmin, CVector3 &out_bboxmax, CVector3 &out_bboxhalfsize, bool useInputTransform, CTransformation in_ct )
{
	double nbox[3];
	double nboxs[3];
											
	PolygonMesh p(in_obj.GetActivePrimitive().GetGeometry());
	CTransformation ct;
	if (useInputTransform)
		ct = in_ct;
	else
		ct = in_obj.GetKinematics().GetGlobal().GetTransform();
		
	//get bounding box
	p.GetBoundingBox(nbox[0],nbox[1],nbox[2],nboxs[0],nboxs[1],nboxs[2],ct);

	nboxs[0]*=0.5;
	nboxs[1]*=0.5;
	nboxs[2]*=0.5;
	//shrink bounding box with preferences value
	//fill array
	out_bboxmin = CVector3(nbox[0]-nboxs[0],nbox[1]-nboxs[1],nbox[2]-nboxs[2]) ;
	out_bboxmax = CVector3(nbox[0]+nboxs[0],nbox[1]+nboxs[1],nbox[2]+nboxs[2]) ;
	out_bboxhalfsize = CVector3(nboxs[0],nboxs[1],nboxs[2]) ;

};

const CStatus GetSinglePolyLocationSample3(ToolContext& in_ctxt, X3DObject& in_obj, CVector3 in_raysample, LONG polyInd, CVector3 &out_pos)
{	
	PolygonMesh pmesh( in_obj.GetActivePrimitive().GetGeometry() );
	CTransformation localToGlobal = in_obj.GetKinematics().GetGlobal().GetTransform();
		
	PolygonFace face = pmesh.GetPolygons().GetItem( polyInd );
	CLongArray pfaceInd = face.GetNeighborPolygons().GetIndexArray();
	pfaceInd.Add(polyInd);
		
	if ( pmesh.SetupPointLocatorQueries( siClosestSurfaceRaycastIntersection,&localToGlobal, pfaceInd.GetCount(),pfaceInd.GetArray(), 10 ) != CStatus::OK)
		return CStatus::False;
		
	CVector3 l_rayview;
		
	if(	in_ctxt.WorldToView(in_raysample, l_rayview) != CStatus::OK )
		return CStatus::Fail;
			
	CLine l_ray;
	in_ctxt.GetWorldRay((LONG)l_rayview[0],(LONG)l_rayview[1],l_ray);
	CVector3	rayt = l_ray.GetTangent();
	CVector3	rayp = l_ray.GetOrigin();

	PointLocatorData	locators;
	locators = pmesh.GetRaycastIntersections( 1, (double*)&rayp, (double*)&rayt, siSemiLineIntersection  );

	if ( !locators.IsPtLocatorValid(0) )
		return CStatus::Fail;		
	if (pmesh.EvaluatePositions(locators, -1, 0, (double*)&out_pos) != CStatus::OK)
		return CStatus::Fail;	
		
	out_pos.MulByTransformationInPlace(localToGlobal);
	return CStatus::OK;
}

const CStatus GetLocationSamples(ToolContext& in_ctxt, X3DObject& l_obj, CVector3 in_direction, CVector3Array& in_raysamples, std::vector<double>& out_pos, std::vector<LONG>& out_polyInd)
	{	
		PolygonMesh pmesh( l_obj.GetActivePrimitive().GetGeometry() );
	
		CTransformation localToGlobal = l_obj.GetKinematics().GetGlobal().GetTransform();
		if ( pmesh.SetupPointLocatorQueries(siClosestSurface/*siClosestSurfaceRaycastIntersection*/,&localToGlobal,-1,NULL,in_raysamples.GetCount()*10 ) != CStatus::OK)		{
			Application().LogMessage(L"Could not setup point locators");
			return CStatus::False;
		}
		PointLocatorData locators;
		
		CVector3Array rays;
		CVector3 dir = in_direction;
		dir.NegateInPlace();
		
		//dir.MulByTransformationInPlace(localToGlobal.InvertInPlace());

		for(int i=0;i<in_raysamples.GetCount();i++)
			rays.Add(dir );
		locators = pmesh.GetClosestLocations(in_raysamples.GetCount(),	(double*)&in_raysamples[0] );	
	//	locators = pmesh.GetRaycastIntersections(in_raysamples.GetCount(),  (double*)&in_raysamples[0], (double*)&rays[0],siSemiLineIntersection);
		CLongArray valid_locators;
		for(LONG i=0;i<locators.GetCount();i++)
		{
			if ( locators.IsPtLocatorValid(i) ) 
				valid_locators.Add(i);
		}
		out_pos.resize( valid_locators.GetCount()*3 );
			
		if (pmesh.EvaluatePositions(locators, valid_locators.GetCount(), (LONG*)&valid_locators[0], &out_pos.front()) != CStatus::OK)	{
			out_pos.clear();
			return CStatus::False;
		}
		out_polyInd.resize( valid_locators.GetCount());
		pmesh.GetPolygonIndexArray(locators, valid_locators.GetCount(),(LONG*)&valid_locators[0], &out_polyInd.front());
		
		return CStatus::OK;
	}

const void BuildCircleSamples(ToolContext& in_ctxt, double scale,LONG in_count, CVector3 center, CVector3 rotx,CVector3 roty,CVector3 rotz,CVector3Array& out_samples)
	{
		double rad = scale;
	
		out_samples.Clear();
		srand(  clock()*clock() );	//seed random number
		CTransformation circleT;
		
		circleT.SetPosX(center.GetX());
		circleT.SetPosY(center.GetY());
		circleT.SetPosZ(center.GetZ());
		circleT.SetRotationFromXYZAxes(rotx,roty,rotz);
	
		for(LONG i=0;i<in_count;i++)
		{						
			double randf = 1.0 - (double)rand()/(double)RAND_MAX;
			double a = ( (double)rand()/(double)RAND_MAX )* 2.0*PI;
			double sqrad = sqrt(randf)*rad;
			CVector3 pt(sqrad* cos(a),sqrad* sin(a), 0.0);
			
			pt.MulByTransformationInPlace(circleT);

			out_samples.Add(pt);
		}			
	}

// ******************************************************************************************************************************************* GET SNAP DATA

const X3DObject GetMultipleComponents(ToolContext& in_ctxt,CRef &target, CRefArray& out_polys)
{			
		X3DObject l_meshObject;
		CComAPIHandler cCollItem(target);
		CComAPIHandler cSubComp(cCollItem.GetProperty(L"SubComponent"));
		CComAPIHandler cCompColl(cSubComp.GetProperty(L"ComponentCollection"));
		CValueArray indices = cSubComp.GetProperty( L"ElementArray" );
		l_meshObject = cSubComp.GetProperty( L"Parent3DObject" ); 
			
		PolygonMesh pmesh( l_meshObject.GetActivePrimitive().GetGeometry() );
		CString l_subcomp = cSubComp.GetProperty(L"Type");
		CPolygonFaceRefArray polyarray = pmesh.GetPolygons();
		if ( l_subcomp == L"polySubComponent" ) 
			for (LONG i=0;i < LONG(indices.GetCount()); i++ )
				out_polys.Add( polyarray.GetItem((LONG)indices[i]) );
			
		return l_meshObject;
}

const CRefArray ExpandSelection( ToolContext& in_ctxt, CRefArray sel)		//convert subcomponent selection to objects
{	
	CRefArray	l_objects; 		
	for (int i=0;i < sel.GetCount();i++)
	{		
		X3DObject obj;
		if (!sel[i].IsA( siX3DObjectID ) )
		{
			CComAPIHandler comCollItem( sel.GetItem(i) );
			CComAPIHandler comSubComp( comCollItem.GetProperty(L"SubComponent") );		
			obj = comSubComp.GetProperty( L"Parent3DObject" );
		}
		else 
			obj = sel[i];
		if (obj.IsA( siX3DObjectID ) ) 					
			l_objects.Add(obj);					
	}
	return l_objects;
}

const CStatus GetSinglePolyLocationSample2(ToolContext& in_ctxt, X3DObject& in_obj, CVector3 in_raysample, LONG polyInd, CVector3 &out_pos, LONG& out_polyInd)
{	

	PolygonMesh pmesh( in_obj.GetActivePrimitive().GetGeometry() );
	CTransformation localToGlobal = in_obj.GetKinematics().GetGlobal().GetTransform();
		
	PolygonFace face = pmesh.GetPolygons().GetItem(  polyInd );
	CLongArray pfaceInd = face.GetNeighborPolygons().GetIndexArray();
	pfaceInd.Add(polyInd);
		
	if ( pmesh.SetupPointLocatorQueries(/*siClosestSurface*/ siClosestSurfaceRaycastIntersection,&localToGlobal, pfaceInd.GetCount(),pfaceInd.GetArray(), 10 ) != CStatus::OK)

		return CStatus::False;
		
	PointLocatorData locators;

	CVector3 l_rayview;
		
	if(	in_ctxt.WorldToView(in_raysample, l_rayview) != CStatus::OK )
		return CStatus::Fail;
		
	CLine l_ray;
	in_ctxt.GetWorldRay((LONG)l_rayview[0],(LONG)l_rayview[1],l_ray);
		
		
//	locators = pmesh.GetClosestLocations(1,	(double*)&in_raysample );	
	CVector3 rayt=l_ray.GetTangent();
	CVector3 rayp=l_ray.GetOrigin();
	locators = pmesh.GetRaycastIntersections( 1, (double*)&rayp, (double*)&rayt, siSemiLineIntersection  );

	if ( !locators.IsPtLocatorValid(0) )
		return CStatus::Fail;
	
	if (pmesh.EvaluatePositions(locators, -1, 0, (double*)&out_pos) != CStatus::OK)
		return CStatus::Fail;	
	
	out_pos.MulByTransformationInPlace(localToGlobal);
	if (pmesh.GetPolygonIndexArray(locators, -1, 0, &out_polyInd)  != CStatus::OK)
		return CStatus::Fail;
	
	return CStatus::OK;
}